use NQPHLL;

class TinyLang::Grammar is HLL::Grammar {
    rule TOP {
        <.newblock>
        <statement> ** ';'
    }
    token newblock { <?> }
    
    proto token statement { <...> }
    token statement:sym<EXPR> { <EXPR> }
    rule statement:sym<class> {
        'a' <class_name>
        '{'
        <class_body_item>*
        '}'
    }
    token class_name { <ident> }
    
    proto token class_body_item { <...> }
    rule class_body_item:sym<method> {
        'can' <ident>
        <.newblock>
        '{' <statement> ** ';' '}'
    }
    
    proto token term { <...> }
    token term:sym<var> { <variable> <![(]> }
    token term:sym<decl> { <variable_declaration> }
    token term:sym<call> { <ident> '(' <arglist> ')' }
    token term:sym<value> { <value> }
    
    rule arglist { <EXPR> ** ',' }
    
    rule variable_declaration { 'var' <variable> }
    token variable { <ident> }
    
    proto token value { <...> }
    token value:sym<string> { <?["]> <quote_EXPR> }
    
    ## Operators
    INIT {
        # Precedence levels.
        TinyLang::Grammar.O(':prec<z=>, :assoc<unary>', '%methodop');
        TinyLang::Grammar.O(':prec<y=>, :assoc<left>',  '%concatenation');
        TinyLang::Grammar.O(':prec<x=>, :assoc<right>', '%assignment');
    }
    token postfix:sym<.> { <sym> <ident> <O('%methodop')> }
    token infix:sym<~>   { <sym> <O('%concatenation')> }
    token infix:sym<=>   { <sym> <O('%assignment')> }
}

class TinyLang::Actions is HLL::Actions {
    my @BLOCK;
    my $CLASS_NAME;
    
    method TOP($/) {
        my $program := @BLOCK.shift;
        for $<statement> {
            $program.push($_.ast);
        }
        make $program;
    }
    
    method statement:sym<EXPR>($/) {
        make $<EXPR>.ast;
    }
    
    method statement:sym<class>($/) {
        # Create type object and put in a lexical.
        @BLOCK[0].symbol($CLASS_NAME, :scope('lexical'));
        my $class_body := PAST::Stmts.new(
            PAST::Op.new(
                :pasttype('bind'),
                PAST::Var.new( :name($CLASS_NAME), :scope('lexical'), :isdecl(1) ),
                PAST::Op.new(
                    :pasttype('callmethod'), :name('new_type'),
                    PAST::Var.new( :name('TinyLangClass'), :scope('contextual') ),
                    PAST::Val.new( :value($CLASS_NAME), :named('name') )
                )
            )
        );
        
        # Add body items.
        for $<class_body_item> {
            $class_body.push($_.ast);
        }
        
        # Compose the class (the "we're done" hook).
        $class_body.push(PAST::Op.new(
            :pasttype('callmethod'), :name('compose'),
            PAST::Op.new(
                :pirop('get_how PP'), PAST::Var.new( :name($CLASS_NAME) )
            ),
            PAST::Var.new( :name($CLASS_NAME) )
        ));
        
        make $class_body;
    }
    
    method class_name($/) {
        $CLASS_NAME := ~$<ident>;
    }
    
    method class_body_item:sym<method>($/) {
        # Add statements to the block.
        my $body := @BLOCK.shift();
        for $<statement> {
            $body.push($_.ast);
        }
        
        # Call add_method on the meta-object.
        make PAST::Op.new(
            :pasttype('callmethod'), :name('add_method'),
            PAST::Op.new(
                :pirop('get_how PP'), PAST::Var.new( :name($CLASS_NAME) )
            ),
            PAST::Var.new( :name($CLASS_NAME) ),
            PAST::Val.new( :value(~$<ident>) ),
            $body
        );
    }
    
    method newblock($/) {
        @BLOCK.unshift(PAST::Block.new());
    }
    
    method term:sym<var>($/) {
        make $<variable>.ast;
    }
    
    method term:sym<decl>($/) {
        make $<variable_declaration>.ast;
    }
    
    method term:sym<call>($/) {
        my $call := $<arglist>.ast;
        $call.pasttype('call');
        $call.name(~$<ident>);
        make $call;
    }
    
    method term:sym<value>($/) {
        make $<value>.ast;
    }
    
    method arglist($/) {
        my $args := PAST::Op.new();
        for $<EXPR> {
            $args.push($_.ast);
        }
        make $args;
    }
    
    method variable_declaration($/) {
        # Put in symbol table as well as making var node.
        my $name := ~$<variable>;
        @BLOCK[0].symbol($name, :scope('lexical'));
        make PAST::Var.new( :name($name), :isdecl(1) );
    }
    
    method variable($/) {
        make PAST::Var.new( :name(~$<ident>) );
    }
    
    method value:sym<string>($/) {
        make $<quote_EXPR>.ast;
    }
    
    method postfix:sym<.>($/) {
        make PAST::Op.new( :pasttype('callmethod'), :name(~$<ident>) );
    }
    
    method infix:sym<~>($/) {
        make PAST::Op.new( :pirop('concat Sss') );
    }
    
    method infix:sym<=>($/) {
        make PAST::Op.new( :pasttype('bind') );
    }
}

class TinyLang::Compiler is HLL::Compiler {
}

my class TinyLangClass {
    has $!name;
    has %!methods;
    
    method new_type(:$name = '<anon>') {
        my $metaclass := self.new(:name($name));
        return pir::repr_type_object_for__PPs($metaclass, 'HashAttrStore');
    }
    
    method add_method($obj, $name, $code) {
        %!methods{$name} := $code;
    }
    
    method compose($obj) {
        return $obj;
    }
    
    method name($obj) {
        return $!name;
    }
    
    method find_method($obj, $name) {
        return %!methods{$name}
    }
}

sub MAIN(*@ARGS) {
    # Create and configure compiler object.
    my $tlcomp := TinyLang::Compiler.new();
    $tlcomp.language('tinylang');
    $tlcomp.parsegrammar(TinyLang::Grammar);
    $tlcomp.parseactions(TinyLang::Actions);

    # Enter the compiler.
    $tlcomp.command_line(@ARGS, :encoding('utf8'));
}

# Built-in subs just come here. :-)
our sub writeline($message) {
    say($message);
}
our sub new($type) {
    return pir::repr_instance_of__PP($type)
}
our sub typeof($obj) {
    return $obj.HOW.name($obj)
}
